/*
 * (C) 2005  Alexander Shaposhnikov  <shaposh@isp.nsc.ru>
 *
 *  Licensed under the terms of the GNU GPL License version 2.
 *  Based upon reverse engineered information
 *
 *  BIG FAT DISCLAIMER: Work in progress code. Possibly *dangerous*
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/cpufreq.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/smp.h>
#include "cpufreq_nForce4.h"

static unsigned int STARTMULT=UNSPECIFIED;
static unsigned int STARTVOLT=UNSPECIFIED;

inline u32 MULT_to_FID(u32 Mult)
{
	u32 fid;
	if(Mult<FID_to_MULT(K8_MIN_FID))
		fid=K8_MIN_FID;
	else
		if(Mult>FID_to_MULT(K8_MAX_FID))
			fid=K8_MAX_FID;
		else
			fid=2*(Mult-4);
	return	fid; 			
}

inline u32 VOLT_to_VID(u32 Volt)
{
	u32 vid;
	if(Volt<VID_to_VOLT(K8_MIN_VID))
		vid=K8_MIN_VID;
	else
		if(Volt>VID_to_VOLT(K8_MAX_VID))
			vid=K8_MAX_VID;
		else
			vid=(1550-Volt)/25;
	return	vid; 			
}



/*Return PLL multiplier from PLL value
*/
static inline u32 PLLMUL(u32 pll)
{
    return   ((pll>>8)&0xfff);

}
/*Return PLL divider from PLL value
*/
static inline u32 PLLDIV(u32 pll)
{
    if(PLLMUL(pll)>=0x200)
	return (pll&0xff)+0x100;
    else
	return (pll&0xff);	 
}
/*Return PLL value from multiplier and divider
 	ToDO: decode entire PLL reg.	
*/
static inline u32 PLLVAL(u32 Mul,u32 Div)
{	
    u32 Pll;	
    Pll=(PLL_FORM)|(Mul<<8);
    Pll=Pll|((Mul>=0x200)?(Div-0x100):Div);
    return Pll;
}
inline void WRITEPLL(u32 PllVal,u16 PllPort)
{
    outl_p(PllVal,PllPort);
    mdelay(PLL_DELAY); 		
}

/*Set Control Register of Specific CPU

*/
int SET_CTL_MSR(struct K8_CTL_MSR Ctl_Msr)
{
	int err=0;	
	u32 eax,edx;
	err=do_rdmsr(Ctl_Msr.cpu,MSR_FIDVID_CTL,&eax,&edx);
	if(err)
	{	
		printk(KERN_ERR "error reading msr %X for CPU %d\n",MSR_FIDVID_CTL,Ctl_Msr.cpu);
		return err;
	}
	eax=eax&(~MSR_C_LO_NEW_FID)&(~MSR_C_LO_NEW_VID);
	eax=eax|Ctl_Msr.Fid;
	eax=eax|(Ctl_Msr.Vid<<8);

	edx=edx&(~MSR_C_HI_STP_GNT_TO);
	edx=edx|Ctl_Msr.Stp_Gnt_To;

	if(Ctl_Msr.Init_Bit)
		eax=eax|(1<<16);	
	err=do_wrmsr(Ctl_Msr.cpu,MSR_FIDVID_CTL,eax,edx); 
	if(err)
	{	
		printk(KERN_ERR "error writing msr %X edx %X eax %X for CPU %d\n",MSR_FIDVID_CTL,edx,eax,Ctl_Msr.cpu);
		return err;
	}
	return err;
}

/*Get current STATUS

*/
int GET_STATUS_MSR(struct K8_STATUS_MSR *Status_Msr)
{
	int err=0;	
	u32 eax,edx;
	err=do_rdmsr(Status_Msr->cpu,MSR_FIDVID_STATUS,&eax,&edx);
	if(err)
	{	
		printk(KERN_ERR "error reading msr %X for CPU %d\n",MSR_FIDVID_STATUS,Status_Msr->cpu);
		return err;
	}
	Status_Msr->CurrFid=eax&MSR_S_LO_CURRENT_FID;
	Status_Msr->StartFid=(eax&MSR_S_LO_START_FID)>>8;
	Status_Msr->MaxFid=(eax&MSR_S_LO_MAX_FID)>>16;
	Status_Msr->MaxRampVid=(eax&MSR_S_LO_MAX_RAMP_VID)>>24;
	Status_Msr->FidVidPnd=(eax&MSR_S_LO_CHANGE_PENDING)>>31;

	Status_Msr->CurrVid=edx&MSR_S_HI_CURRENT_VID;	
	Status_Msr->StartVid=(edx&MSR_S_HI_START_VID)>>8;
	Status_Msr->MaxVid=(edx&MSR_S_HI_MAX_WORKING_VID)>>16;
	Status_Msr->MinVid=(edx&MSR_S_HI_MIN_WORKING_VID)>>24;
	if(Status_Msr->MinVid<Status_Msr->MaxVid)
		Status_Msr->MinVid=K8_MIN_VID;
	return err;  		   				

}
/*Returns current FSB is KHz.Until i could find how to read current FSB regardless 
	of PLL value, this function can return wrong value if PLL_ENABLE is not set
*/
static int GET_FSB(u32 *Fsb)
{	
	int ret=0;
	u16 tmp;
	u32 mul,div;    
	u32 Pll_Port;
	u32 Pll_Enable_Port;
	u32 Pll_Val;
	u32 Pll_Enable_Val;
	pci_read_config_word(CK804_chipset_dev, PLL_ADR, &tmp);			
	
	Pll_Port=(tmp&0xff00)+0x8;
	Pll_Enable_Port=(tmp&0xff00)+0xC8;

	Pll_Enable_Val=inl_p(Pll_Enable_Port);
	/*If PLL_ENABLE is not zero, PLL value is probably wrong 
		Pll_Enable_Val: 0-enable,1-disable
	*/
	if(Pll_Enable_Val&0x1)
	{
		printk(KERN_WARNING "cpufreq:PLL is disabled. FSB reading may be inaccurate!\n");
		ret=-1;
	}

	Pll_Val=inl_p(Pll_Port);
	
	mul=PLLMUL(Pll_Val);
	div=PLLDIV(Pll_Val);
	*Fsb=PLL_CLOCK*mul/div;
	
	return ret;
} 

/*Set FSB. PLL divider is set to 0x100(0x00) and
	PLL multiplier is varied by steps of PLL_MUL_STEP
*/
int SET_FSB(u32 Target_Fsb)
{	
	int ret=0;
	u32 Curr_Fsb;
	u16 tmp;
	u16 Pll_Port,Pll_Enable_Port;
	u32 Pll_Val;
	u32 mul,div;
	u32 Pll_Enable_Val;
	pci_read_config_word(CK804_chipset_dev, PLL_ADR, &tmp);			
	
	Pll_Port=(tmp&0xff00)+0x8;
	Pll_Enable_Port=(tmp&0xff00)+0xC8;

	Pll_Val=inl_p(Pll_Port);
	Pll_Enable_Val=inl_p(Pll_Enable_Port);

	
	/*Set PLL_ENABLE to 0=enable. May crash here!*/
	if(Pll_Enable_Val&0x1)
	{
		outl_p(Pll_Enable_Val&(~0x1),Pll_Enable_Port);
		mdelay(PLL_DELAY); 
	}
	mul=PLLMUL(Pll_Val);
	div=PLLDIV(Pll_Val);

	/*I call it type 1 values. Probably there are more "types". We must jump to
		corresponding type2 , with mul>=0x200.
	 */
	if(mul<0x200) 
	{

		mul=PLL_DIV*mul/div;
		outl_p(PLLVAL(mul,PLL_DIV),Pll_Port); 
		mdelay(PLL_DELAY); 

	}
	Curr_Fsb=PLL_CLOCK*mul/PLL_DIV;
	
	if(Curr_Fsb==Target_Fsb)
	    return 0;
	    
	/*This algorithm should always set Final_Fsb<=Target_Fsb*/    
	if(Curr_Fsb<Target_Fsb)
	{
	    mul+=PLL_MUL_STEP;		
	    Curr_Fsb=PLL_CLOCK*mul/PLL_DIV;
	    for(;Curr_Fsb<=Target_Fsb;)				
	    {	
		WRITEPLL(PLLVAL(mul,PLL_DIV),Pll_Port);
		mul+=PLL_MUL_STEP;		
		Curr_Fsb=PLL_CLOCK*mul/PLL_DIV;
	    }
	    /*Fine tuning*/
	    mul-=PLL_MUL_STEP;
	    Curr_Fsb=PLL_CLOCK*mul/PLL_DIV;	
	    if(Curr_Fsb<Target_Fsb)
	    {	
		mul++;		
		Curr_Fsb=PLL_CLOCK*mul/PLL_DIV;		
		for(;Curr_Fsb<=Target_Fsb;)
		{
		    WRITEPLL(PLLVAL(mul,PLL_DIV),Pll_Port);		
		    mul++;
		    Curr_Fsb=PLL_CLOCK*mul/PLL_DIV;	
		}    
	    }	
	}    
    	else /*Curr_Fsb>Target_Fsb*/
	{
	    for(;Curr_Fsb>Target_Fsb;)				
	    {	
		mul-=PLL_MUL_STEP;		
		Curr_Fsb=PLL_CLOCK*mul/PLL_DIV;
		WRITEPLL(PLLVAL(mul,PLL_DIV),Pll_Port);		
	    }
	    /*Fine tuning*/
	    if(Curr_Fsb<Target_Fsb)
	    {	
		mul++;		
		Curr_Fsb=PLL_CLOCK*mul/PLL_DIV;		
		for(;Curr_Fsb<=Target_Fsb;)
		{
		    WRITEPLL(PLLVAL(mul,PLL_DIV),Pll_Port);		
		    mul++;
		    Curr_Fsb=PLL_CLOCK*mul/PLL_DIV;	
		}    
	    }
	}    
	return ret;
} 

static int CK804_target(struct cpufreq_policy *policy,
			  unsigned int target_freq, unsigned int relation)
{
	u32 target_fsb;
	u32 current_fsb;
	u32 old_fsb;
	int i;
	struct cpufreq_freqs freqs[MAX_CPU_NUMBER];
	struct K8_STATUS_MSR Status_Msr[MAX_CPU_NUMBER];
	
	
	if ((target_freq > policy->max) || (target_freq < policy->min))
		return -EINVAL;

/*We have to send notifiers with all CPUs!*/

	for(i=0;i<NR_CPUS;i++)
	{
		Status_Msr[i].cpu=i;
		GET_STATUS_MSR(&(Status_Msr[i]));
	}

	target_fsb = target_freq/(FID_to_MULT(Status_Msr[policy->cpu].CurrFid));
	
	GET_FSB(&current_fsb);
	old_fsb=current_fsb;

	for(i=0;i<NR_CPUS;i++)
	{	
		u32 mult;
		mult=FID_to_MULT(Status_Msr[i].CurrFid);
		freqs[i].old = current_fsb*mult;
		freqs[i].new = target_fsb*mult;
		freqs[i].cpu = i;
	}		

	if (freqs[policy->cpu].old == freqs[policy->cpu].new)
		return 0;
	
	for(i=0;i<NR_CPUS;i++)
		cpufreq_notify_transition(&(freqs[i]), CPUFREQ_PRECHANGE);

	// Disable IRQs ?

	if (SET_FSB(target_fsb) < 0)
	{
	    printk(KERN_ERR "cpufreq: FSB changing from %d to %d failed\n",current_fsb,target_fsb);
	    	       
	}
	else
	{
	    GET_FSB(&current_fsb);	
	    printk(KERN_INFO "cpufreq: FSB changed successfully from %d to %d\n",old_fsb,current_fsb);	    		
	}
	    	       
	 //Enable IRQs ?
	
	for(i=0;i<NR_CPUS;i++)
	{
	    freqs[i].new = current_fsb*FID_to_MULT(Status_Msr[i].CurrFid);	
	    cpufreq_notify_transition(&(freqs[i]), CPUFREQ_POSTCHANGE);

        }
	

	return 0;
}
static int CK804_verify(struct cpufreq_policy *policy)
{
	struct K8_STATUS_MSR Status_Msr;
	u32 curfsb;
	u32 min,max;    

	Status_Msr.cpu=policy->cpu;
	GET_STATUS_MSR(&Status_Msr);

	GET_FSB(&curfsb);
	min = PLL_MIN_FSB*FID_to_MULT(Status_Msr.CurrFid);
	max = PLL_MAX_FSB*FID_to_MULT(Status_Msr.CurrFid);
	
	
	cpufreq_verify_within_limits(policy,min,max);

	return 0;
}
static unsigned int CK804_get(unsigned int cpu)
{
	u32 curfsb;
	u32 curclock;    
	struct K8_STATUS_MSR Status_Msr;

	GET_FSB(&curfsb);

	Status_Msr.cpu=cpu;
	GET_STATUS_MSR(&Status_Msr);
	curclock=curfsb*FID_to_MULT(Status_Msr.CurrFid);
	
	return curclock ;
}
static int CK804_cpu_init(struct cpufreq_policy *policy)
{
	struct K8_STATUS_MSR Status_Msr;
	u32 curfsb;

	Status_Msr.cpu=policy->cpu;
	GET_STATUS_MSR(&Status_Msr);

	GET_FSB(&curfsb);

	policy->cpuinfo.min_freq = PLL_MIN_FSB*FID_to_MULT(K8_MIN_FID);
	policy->cpuinfo.max_freq = PLL_MAX_FSB*FID_to_MULT(Status_Msr.MaxFid);

	policy->cpuinfo.transition_latency = CPUFREQ_ETERNAL;
	policy->cur = curfsb*FID_to_MULT(Status_Msr.CurrFid);
	if( (curfsb-PLL_SAFE_DISTANCE)>PLL_MIN_FSB )
		policy->min = (curfsb-PLL_SAFE_DISTANCE)*FID_to_MULT(Status_Msr.CurrFid);	
	else
		policy->min = PLL_MIN_FSB*FID_to_MULT(Status_Msr.CurrFid);	
	    		
	policy->max = curfsb*FID_to_MULT(Status_Msr.CurrFid);
	policy->governor = CPUFREQ_DEFAULT_GOVERNOR;
	
	printk(KERN_INFO "cpufreq: on the start cpu %d,Voltage=%d,Multiplier=%d\n", policy->cpu,VID_to_VOLT(Status_Msr.CurrVid),FID_to_MULT(Status_Msr.CurrFid));
	return 0;
}
static int CK804_cpu_exit(struct cpufreq_policy *policy)
{
	return 0;
}

static struct cpufreq_driver CK804_driver = {
	.name   = "cpufreq_nForce4",
	.verify = CK804_verify,
	.target = CK804_target,
	.get    = CK804_get,
	.init   = CK804_cpu_init,
	.exit   = CK804_cpu_exit,
	.owner  = THIS_MODULE,
};


/**
 * nForce4_detect_chipset - detect the Southbridge which contains FSB PLL logic
 *
 * Detects CK804 revision
 * 
 */
static unsigned int CK804_detect_chipset(void)
{
	u8 revision;

	CK804_chipset_dev = pci_get_subsys(PCI_VENDOR_ID_NVIDIA,
                                             PCI_DEVICE_ID_NVIDIA_CK804_PRO,
                                             PCI_ANY_ID,
                                             PCI_ANY_ID,
                                             NULL);

	if (CK804_chipset_dev == NULL)
		return -ENODEV;

	pci_read_config_byte(CK804_chipset_dev, PCI_REVISION_ID, &revision);

	printk(KERN_INFO "cpufreq: Detected CK804 chipset revision %X\n",
	       revision);
	printk(KERN_INFO "cpufreq: FSB changing is dangerous and can lead to crashes and data loss\n");

	return 0;
}


/**
 * CK804_init - initializes the CK804 CPUFreq driver
 *
 */
static int __init CK804_init(void)
{	
	int i;
	
	/* detect chipset */
	if (CK804_detect_chipset()) {
		printk(KERN_ERR "cpufreq: No CK804 chipset.\n");
		return -ENODEV;
	}
	/*TODO: fix this, have to send notifiers on FID change
	    
	*/
		
	for(i=0;i<NR_CPUS;i++)
	{
		struct K8_STATUS_MSR Status_Msr;
		struct K8_CTL_MSR Ctl_Msr;

		Status_Msr.cpu=i;			
		GET_STATUS_MSR(&Status_Msr);		
	
		if((STARTVOLT!=UNSPECIFIED)&&(Status_Msr.CurrVid!=VOLT_to_VID(STARTVOLT)))
		{
			/*Need to set VID first*/		
			Ctl_Msr.cpu=i;
			Ctl_Msr.Fid=Status_Msr.CurrFid;
			Ctl_Msr.Vid=VOLT_to_VID(STARTVOLT);
			Ctl_Msr.Stp_Gnt_To=DEF_STP_GNT_TO;
			Ctl_Msr.Init_Bit=1;
			SET_CTL_MSR(Ctl_Msr);			
		    	GET_STATUS_MSR(&Status_Msr);	
		}	
		if((STARTMULT!=UNSPECIFIED)&&(Status_Msr.CurrFid!=MULT_to_FID(STARTMULT)))
		{
			/*Need to set FID */
			Ctl_Msr.cpu=i;
			Ctl_Msr.Vid=Status_Msr.CurrVid;
			Ctl_Msr.Fid=MULT_to_FID(STARTMULT);		
			Ctl_Msr.Stp_Gnt_To=DEF_STP_GNT_TO;
			Ctl_Msr.Init_Bit=1;	    		
			SET_CTL_MSR(Ctl_Msr);
    		}
	}
	return cpufreq_register_driver(&CK804_driver);
	
}

/**
 * CK804_exit - unregisters cpufreq module
 *
 *   Unregisters CK804 FSB change support.
 */
static void __exit CK804_exit(void)
{	
	printk(KERN_INFO "cpufreq: Exiting\n");	
	cpufreq_unregister_driver(&CK804_driver); 
}

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Alexander Shaposhnikov <shaposh@isp.nsc.ru>");
MODULE_DESCRIPTION("nVidia CK804 (nForce4) FSB changing cpufreq driver ver 0.1");
module_param(STARTMULT,int,0444);
module_param(STARTVOLT,int,0444);


module_init(CK804_init);
module_exit(CK804_exit);
