/* Shamelessly ripped from "msr.c" kernel driver
*/
/* ----------------------------------------------------------------------- *
 *   
 *   Copyright 2000 H. Peter Anvin - All Rights Reserved
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, Inc., 675 Mass Ave, Cambridge MA 02139,
 *   USA; either version 2 of the License, or (at your option) any later
 *   version; incorporated herein by reference.
 *
 * ----------------------------------------------------------------------- */
#include <asm/msr.h>

/* Note: "err" is handled in a funny way below.  Otherwise one version
   of gcc or another breaks. */
static inline int wrmsr_eio(u32 reg, u32 eax, u32 edx)
{
	int err;

	asm volatile ("1:	wrmsr\n"
		      "2:\n"
		      ".section .fixup,\"ax\"\n"
		      "3:	movl %4,%0\n"
		      "	jmp 2b\n"
		      ".previous\n"
		      ".section __ex_table,\"a\"\n"
		      "	.align 8\n" "	.quad 1b,3b\n" ".previous":"=&bDS" (err)
		      :"a"(eax), "d"(edx), "c"(reg), "i"(-EIO), "0"(0));

	return err;
}
static inline int rdmsr_eio(u32 reg, u32 *eax, u32 *edx)
{
	int err;

	asm volatile ("1:	rdmsr\n"
		      "2:\n"
		      ".section .fixup,\"ax\"\n"
		      "3:	movl %4,%0\n"
		      "	jmp 2b\n"
		      ".previous\n"
		      ".section __ex_table,\"a\"\n"
		      "	.align 8\n"
		      "	.quad 1b,3b\n"
		      ".previous":"=&bDS" (err), "=a"(*eax), "=d"(*edx)
		      :"c"(reg), "i"(-EIO), "0"(0));

	return err;
}
#ifdef CONFIG_SMP
struct msr_command {
	int cpu;
	int err;
	u32 reg;
	u32 data[2];
};
static void msr_smp_wrmsr(void *cmd_block)
{
	struct msr_command *cmd = (struct msr_command *)cmd_block;

	if (cmd->cpu == smp_processor_id())
		cmd->err = wrmsr_eio(cmd->reg, cmd->data[0], cmd->data[1]);
}
static void msr_smp_rdmsr(void *cmd_block)
{
	struct msr_command *cmd = (struct msr_command *)cmd_block;

	if (cmd->cpu == smp_processor_id())
		cmd->err = rdmsr_eio(cmd->reg, &cmd->data[0], &cmd->data[1]);
}
static int do_wrmsr(int cpu, u32 reg, u32 eax, u32 edx)
{
	struct msr_command cmd;
	int ret;

	preempt_disable();
	if (cpu == smp_processor_id()) {
		ret = wrmsr_eio(reg, eax, edx);
	} else {
		cmd.cpu = cpu;
		cmd.reg = reg;
		cmd.data[0] = eax;
		cmd.data[1] = edx;

		smp_call_function(msr_smp_wrmsr, &cmd, 1, 1);
		ret = cmd.err;
	}
	preempt_enable();
	return ret;
}
static  int do_rdmsr(int cpu, u32 reg, u32 * eax, u32 * edx)
{
	struct msr_command cmd;
	int ret;

	preempt_disable();
	if (cpu == smp_processor_id()) {
		ret = rdmsr_eio(reg, eax, edx);
	} else {
		cmd.cpu = cpu;
		cmd.reg = reg;

		smp_call_function(msr_smp_rdmsr, &cmd, 1, 1);

		*eax = cmd.data[0];
		*edx = cmd.data[1];

		ret = cmd.err;
	}
	preempt_enable();
	return ret;
}
#else				/* ! CONFIG_SMP */
static int do_wrmsr(int cpu, u32 reg, u32 eax, u32 edx)
{
	return wrmsr_eio(reg, eax, edx);
}
static int do_rdmsr(int cpu, u32 reg, u32 *eax, u32 *edx)
{
	return rdmsr_eio(reg, eax, edx);
}
#endif				/* ! CONFIG_SMP */
/*----------End of msr.c kernel driver-------------*/

#define DRV_VERSION 	    "0.1"			
#define UNSPECIFIED         0xffffffff
#define MAX_CPU_NUMBER	    16
#define PLL_MIN_FSB	    100000
#define PLL_MAX_FSB	    400000		
#define PLL_DEFAULT_FSB     200000
#define PLL_SAFE_DISTANCE    50000

/*CK804 PLL related items 
  TODO: find value of PLL_CLOCK more precisely	
*/
#define PLL_ADR		    0x68	
#define PLL_CLOCK 	    48438
#define PLL_MUL_STEP        4		
#define PLL_DIV             0x100 
#define PLL_DELAY           5
#define PLL_FORM            0x81E00000
/* Model Specific Registers. MSRs are 64-bit. For     */
/* writes (wrmsr - opcode 0f 30), the register number is placed in ecx, and   */
/* the value to write is placed in edx:eax. For reads (rdmsr - opcode 0f 32), */
/* the register number is placed in ecx, and the data is returned in edx:eax. */
#define MSR_FIDVID_CTL      0xc0010041
#define MSR_FIDVID_STATUS   0xc0010042
#define K8_MIN_VID	    0x1E
#define K8_MAX_VID          0
#define K8_MAX_FID	    0x2A
#define K8_MIN_FID	    0
#define DEF_STP_GNT_TO	    0x800			
/* Field definitions within the FID VID Low Control MSR : */
#define MSR_C_LO_INIT_FID_VID     0x00010000
#define MSR_C_LO_NEW_VID          0x00001f00
#define MSR_C_LO_NEW_FID          0x0000003f
/* Field definitions within the FID VID High Control MSR : */
#define MSR_C_HI_STP_GNT_TO 	  0x000fffff
/* Field definitions within the FID VID Low Status MSR : */
#define MSR_S_LO_CHANGE_PENDING   0x80000000   /* cleared when completed */
#define MSR_S_LO_MAX_RAMP_VID     0x1f000000
#define MSR_S_LO_MAX_FID          0x003f0000
#define MSR_S_LO_START_FID        0x00003f00
#define MSR_S_LO_CURRENT_FID      0x0000003f
/* Field definitions within the FID VID High Status MSR : */
#define MSR_S_HI_MIN_WORKING_VID  0x1f000000
#define MSR_S_HI_MAX_WORKING_VID  0x001f0000
#define MSR_S_HI_START_VID        0x00001f00
#define MSR_S_HI_CURRENT_VID      0x0000001f
/*----------------------------*/
/*Conversions*/
#define FID_to_MULT(FID)		(FID/2+4) 
#define VID_to_VOLT(VID)		(1550-25*VID) 

/*CK804 ISA Bridge*/
static struct pci_dev *CK804_chipset_dev;

struct K8_STATUS_MSR
{
	int cpu;
	u32 CurrFid;
	u32 StartFid;
	u32 MaxFid;
	u32 MaxRampVid;
	u32 FidVidPnd;
	u32 CurrVid;
	u32 StartVid;
	u32 MaxVid;
	u32 MinVid;	
};
struct K8_CTL_MSR
{
	int cpu;
	u32 Fid;
	u32 Vid;
	u32 Stp_Gnt_To;
	u32 Init_Bit;
};
